import inspect
import logging
from abc import ABC, abstractmethod
from collections.abc import MutableMapping
from typing import IO, TYPE_CHECKING, Any, Generic, Literal, cast

import torch
import torch.distributed
from lightning.fabric.utilities.types import _MAP_LOCATION_TYPE, _PATH
from lightning.pytorch import LightningModule
from lightning.pytorch.profilers import PassThroughProfiler, Profiler
from lightning.pytorch.utilities.types import STEP_OUTPUT
from typing_extensions import Self, TypeVar, override

from ..callbacks.rlp_sanity_checks import _RLPSanityCheckModuleMixin
from ..util._environment_info import EnvironmentConfig
from .config import BaseConfig
from .mixins.callback import CallbackModuleMixin
from .mixins.logger import LoggerLightningModuleMixin

log = logging.getLogger(__name__)

THparams = TypeVar("THparams", bound=BaseConfig, infer_variance=True)


T = TypeVar("T", infer_variance=True)

ReduceOpStr = Literal[
    "avg",
    "mean",
    "band",
    "bor",
    "bxor",
    "max",
    "min",
    "premul_sum",
    "product",
    "sum",
]
VALID_REDUCE_OPS = (
    "avg",
    "mean",
    "band",
    "bor",
    "bxor",
    "max",
    "min",
    "premul_sum",
    "product",
    "sum",
)


class LightningModuleBase(  # pyright: ignore[reportIncompatibleMethodOverride]
    _RLPSanityCheckModuleMixin,
    LoggerLightningModuleMixin,
    CallbackModuleMixin,
    LightningModule,
    ABC,
    Generic[THparams],
):
    # region Config
    @torch.jit.unused
    @property
    def config(self) -> THparams:
        return self.hparams

    @property
    def debug(self) -> bool:
        if torch.jit.is_scripting():
            return False
        return self.config.debug

    # endregion

    # region Debug

    @torch.jit.unused
    def breakpoint(self, rank_zero_only: bool = True):
        if (
            not rank_zero_only
            or not torch.distributed.is_initialized()
            or torch.distributed.get_rank() == 0
        ):
            breakpoint()

        if rank_zero_only and torch.distributed.is_initialized():
            _ = torch.distributed.barrier()

    @torch.jit.unused
    def ensure_finite(
        self,
        tensor: torch.Tensor,
        name: str | None = None,
        throw: bool = False,
    ):
        name_parts: list[str] = ["Tensor"]
        if name is not None:
            name_parts.append(name)
        name = " ".join(name_parts)

        not_finite = ~torch.isfinite(tensor)
        if not_finite.any():
            msg = f"{name} has {not_finite.sum().item()}/{not_finite.numel()} non-finite values."
            if throw:
                raise RuntimeError(msg)
            else:
                log.warning(msg)
            return False
        return True

    # endregion

    # region Profiler
    @property
    def profiler(self) -> Profiler:
        if (trainer := self._trainer) is None:
            raise RuntimeError("trainer is not defined")

        if not hasattr(trainer, "profiler"):
            raise RuntimeError("trainer does not have profiler")

        if (profiler := getattr(trainer, "profiler")) is None:
            profiler = PassThroughProfiler()

        return profiler

    # endregion

    # region Distributed
    def all_gather_object(
        self,
        object: T,
        group: torch.distributed.ProcessGroup | None = None,
    ) -> list[T]:
        if (
            not torch.distributed.is_available()
            or not torch.distributed.is_initialized()
        ):
            return [object]

        object_list = [cast(T, None) for _ in range(self.trainer.world_size)]
        torch.distributed.all_gather_object(object_list, object, group=group)
        return object_list

    def barrier(self, name: str | None = None):
        self.trainer.strategy.barrier(name=name)

    def reduce(
        self,
        tensor: torch.Tensor,
        reduce_op: torch.distributed.ReduceOp.RedOpType | ReduceOpStr,
        group: Any | None = None,
    ) -> torch.Tensor:
        if isinstance(reduce_op, str):
            # validate reduce_op
            if reduce_op not in VALID_REDUCE_OPS:
                raise ValueError(
                    f"reduce_op must be one of {VALID_REDUCE_OPS}, got {reduce_op}"
                )

        return self.trainer.strategy.reduce(tensor, group=group, reduce_op=reduce_op)

    # endregion

    # Our own custom __repr__ method.
    # Torch's __repr__ method is too verbose and doesn't provide any useful information.
    @override
    def __repr__(self):
        parts: list[str] = []
        parts.append(f"config={self.hparams.concise_repr()}")
        parts.append(f"device={self.device}")
        if self.debug:
            parts.append("debug=True")

        parts_str = ", ".join(parts)
        return f"{self.__class__.__name__}({parts_str})"

    @classmethod
    def _validate_class_for_ckpt_loading(cls):
        # Make sure that the `__init__` method takes a single argument, `hparams`.
        if (init_fn := getattr(cls, "__init__", None)) is None:
            return

        if not inspect.isfunction(init_fn):
            raise TypeError(f"__init__ must be a function: {init_fn}")

        parameters = dict(inspect.signature(init_fn).parameters)
        # Remove the "self" parameter.
        _ = parameters.pop("self", None)
        if len(parameters) != 1:
            raise TypeError(
                f"__init__ must take a single argument, got {len(parameters)}: {init_fn}"
            )

        if "hparams" not in parameters:
            raise TypeError(
                f"__init__'s argument must be named 'hparams', got {parameters}"
            )

    hparams: THparams  # pyright: ignore[reportIncompatibleMethodOverride]
    hparams_initial: THparams  # pyright: ignore[reportIncompatibleMethodOverride]

    @classmethod
    @abstractmethod
    def config_cls(cls) -> type[THparams]: ...

    @classmethod
    def load_checkpoint(
        cls,
        checkpoint_path: _PATH | IO,
        hparams: THparams | MutableMapping[str, Any] | None = None,
        map_location: _MAP_LOCATION_TYPE = None,
        strict: bool = True,
    ) -> Self:
        if strict:
            cls._validate_class_for_ckpt_loading()

        kwargs: dict[str, Any] = {}
        if hparams is not None:
            kwargs["hparams"] = hparams

        return super().load_from_checkpoint(
            checkpoint_path,
            map_location=map_location,
            hparams_file=None,
            strict=strict,
            **kwargs,
        )

    def pre_init_update_hparams_dict(self, hparams: MutableMapping[str, Any]):
        """
        Override this method to update the hparams dictionary before it is used to create the hparams object.
        Mapping-based parameters are passed to the constructor of the hparams object when we're loading the model from a checkpoint.
        """
        return hparams

    def pre_init_update_hparams(self, hparams: THparams):
        """
        Override this method to update the hparams object before it is used to create the hparams_initial object.
        """
        return hparams

    @override
    def __init__(self, hparams: THparams | MutableMapping[str, Any]):
        if not isinstance(hparams, BaseConfig):
            if not isinstance(hparams, MutableMapping):
                raise TypeError(
                    f"hparams must be a BaseConfig or a MutableMapping: {type(hparams)}"
                )

            hparams = self.pre_init_update_hparams_dict(hparams)
            hparams = self.config_cls().model_validate(hparams)
        hparams.environment = EnvironmentConfig.from_current_environment(hparams, self)
        hparams = self.pre_init_update_hparams(hparams)

        super().__init__()
        self.save_hyperparameters(hparams)

    def zero_loss(self):
        """
        Returns a loss tensor with the value 0.
        It multiples each weight by 0 and returns the sum, so we don't run into issues with ununsed parameters in DDP.
        """
        loss = sum((0.0 * v).sum() for v in self.parameters() if v.requires_grad)
        loss = cast(torch.Tensor, loss)
        return loss

    if TYPE_CHECKING:

        @override
        def training_step(  # pyright: ignore[reportIncompatibleMethodOverride]
            self,
            batch: Any,
            batch_idx: int,
        ) -> Any:
            r"""Here you compute and return the training loss and some additional metrics for e.g. the progress bar or
            logger.

            Args:
                batch: The output of your data iterable, normally a :class:`~torch.utils.data.DataLoader`.
                batch_idx: The index of this batch.
                dataloader_idx: The index of the dataloader that produced this batch.
                    (only if multiple dataloaders used)

            Return:
                - :class:`~torch.Tensor` - The loss tensor
                - ``dict`` - A dictionary which can include any keys, but must include the key ``'loss'`` in the case of
                automatic optimization.
                - ``None`` - In automatic optimization, this will skip to the next batch (but is not supported for
                multi-GPU, TPU, or DeepSpeed). For manual optimization, this has no special meaning, as returning
                the loss is not required.

            In this step you'd normally do the forward pass and calculate the loss for a batch.
            You can also do fancier things like multiple forward passes or something model specific.

            Example::

                def training_step(self, batch, batch_idx):
                    x, y, z = batch
                    out = self.encoder(x)
                    loss = self.loss(out, x)
                    return loss

            To use multiple optimizers, you can switch to 'manual optimization' and control their stepping:

            .. code-block:: python

                def __init__(self):
                    super().__init__()
                    self.automatic_optimization = False


                # Multiple optimizers (e.g.: GANs)
                def training_step(self, batch, batch_idx):
                    opt1, opt2 = self.optimizers()

                    # do training_step with encoder
                    ...
                    opt1.step()
                    # do training_step with decoder
                    ...
                    opt2.step()

            Note:
                When ``accumulate_grad_batches`` > 1, the loss returned here will be automatically
                normalized by ``accumulate_grad_batches`` internally.

            """
            raise NotImplementedError

        @override
        def validation_step(  # pyright: ignore[reportIncompatibleMethodOverride]
            self,
            batch: Any,
            batch_idx: int,
        ) -> STEP_OUTPUT:
            r"""Operates on a single batch of data from the validation set. In this step you'd might generate examples or
            calculate anything of interest like accuracy.

            Args:
                batch: The output of your data iterable, normally a :class:`~torch.utils.data.DataLoader`.
                batch_idx: The index of this batch.
                dataloader_idx: The index of the dataloader that produced this batch.
                    (only if multiple dataloaders used)

            Return:
                - :class:`~torch.Tensor` - The loss tensor
                - ``dict`` - A dictionary. Can include any keys, but must include the key ``'loss'``.
                - ``None`` - Skip to the next batch.

            .. code-block:: python

                # if you have one val dataloader:
                def validation_step(self, batch, batch_idx): ...


                # if you have multiple val dataloaders:
                def validation_step(self, batch, batch_idx, dataloader_idx=0): ...

            Examples::

                # CASE 1: A single validation dataset
                def validation_step(self, batch, batch_idx):
                    x, y = batch

                    # implement your own
                    out = self(x)
                    loss = self.loss(out, y)

                    # log 6 example images
                    # or generated text... or whatever
                    sample_imgs = x[:6]
                    grid = torchvision.utils.make_grid(sample_imgs)
                    self.logger.experiment.add_image('example_images', grid, 0)

                    # calculate acc
                    labels_hat = torch.argmax(out, dim=1)
                    val_acc = torch.sum(y == labels_hat).item() / (len(y) * 1.0)

                    # log the outputs!
                    self.log_dict({'val_loss': loss, 'val_acc': val_acc})

            If you pass in multiple val dataloaders, :meth:`validation_step` will have an additional argument. We recommend
            setting the default value of 0 so that you can quickly switch between single and multiple dataloaders.

            .. code-block:: python

                # CASE 2: multiple validation dataloaders
                def validation_step(self, batch, batch_idx, dataloader_idx=0):
                    # dataloader_idx tells you which dataset this is.
                    ...

            Note:
                If you don't need to validate you don't need to implement this method.

            Note:
                When the :meth:`validation_step` is called, the model has been put in eval mode
                and PyTorch gradients have been disabled. At the end of validation,
                the model goes back to training mode and gradients are enabled.

            """
            raise NotImplementedError

        @override
        def test_step(  # pyright: ignore[reportIncompatibleMethodOverride]
            self,
            batch: Any,
            batch_idx: int,
        ) -> STEP_OUTPUT:
            r"""Operates on a single batch of data from the test set. In this step you'd normally generate examples or
            calculate anything of interest such as accuracy.

            Args:
                batch: The output of your data iterable, normally a :class:`~torch.utils.data.DataLoader`.
                batch_idx: The index of this batch.
                dataloader_idx: The index of the dataloader that produced this batch.
                    (only if multiple dataloaders used)

            Return:
                - :class:`~torch.Tensor` - The loss tensor
                - ``dict`` - A dictionary. Can include any keys, but must include the key ``'loss'``.
                - ``None`` - Skip to the next batch.

            .. code-block:: python

                # if you have one test dataloader:
                def test_step(self, batch, batch_idx): ...


                # if you have multiple test dataloaders:
                def test_step(self, batch, batch_idx, dataloader_idx=0): ...

            Examples::

                # CASE 1: A single test dataset
                def test_step(self, batch, batch_idx):
                    x, y = batch

                    # implement your own
                    out = self(x)
                    loss = self.loss(out, y)

                    # log 6 example images
                    # or generated text... or whatever
                    sample_imgs = x[:6]
                    grid = torchvision.utils.make_grid(sample_imgs)
                    self.logger.experiment.add_image('example_images', grid, 0)

                    # calculate acc
                    labels_hat = torch.argmax(out, dim=1)
                    test_acc = torch.sum(y == labels_hat).item() / (len(y) * 1.0)

                    # log the outputs!
                    self.log_dict({'test_loss': loss, 'test_acc': test_acc})

            If you pass in multiple test dataloaders, :meth:`test_step` will have an additional argument. We recommend
            setting the default value of 0 so that you can quickly switch between single and multiple dataloaders.

            .. code-block:: python

                # CASE 2: multiple test dataloaders
                def test_step(self, batch, batch_idx, dataloader_idx=0):
                    # dataloader_idx tells you which dataset this is.
                    ...

            Note:
                If you don't need to test you don't need to implement this method.

            Note:
                When the :meth:`test_step` is called, the model has been put in eval mode and
                PyTorch gradients have been disabled. At the end of the test epoch, the model goes back
                to training mode and gradients are enabled.

            """
            raise NotImplementedError

    @override
    def predict_step(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        batch: Any,
        batch_idx: int,
    ) -> STEP_OUTPUT:
        """Step function called during :meth:`~lightning.pytorch.trainer.trainer.Trainer.predict`. By default, it calls
        :meth:`~lightning.pytorch.core.LightningModule.forward`. Override to add any processing logic.

        The :meth:`~lightning.pytorch.core.LightningModule.predict_step` is used
        to scale inference on multi-devices.

        To prevent an OOM error, it is possible to use :class:`~lightning.pytorch.callbacks.BasePredictionWriter`
        callback to write the predictions to disk or database after each batch or on epoch end.

        The :class:`~lightning.pytorch.callbacks.BasePredictionWriter` should be used while using a spawn
        based accelerator. This happens for ``Trainer(strategy="ddp_spawn")``
        or training on 8 TPU cores with ``Trainer(accelerator="tpu", devices=8)`` as predictions won't be returned.

        Args:
            batch: The output of your data iterable, normally a :class:`~torch.utils.data.DataLoader`.
            batch_idx: The index of this batch.
            dataloader_idx: The index of the dataloader that produced this batch.
                (only if multiple dataloaders used)

        Return:
            Predicted output (optional).

        Example ::

            class MyModel(LightningModule):

                def predict_step(self, batch, batch_idx, dataloader_idx=0):
                    return self(batch)

            dm = ...
            model = MyModel()
            trainer = Trainer(accelerator="gpu", devices=2)
            predictions = trainer.predict(model, dm)

        """
        prediction = self(batch)
        return {
            "prediction": prediction,
            "batch": batch,
            "batch_idx": batch_idx,
        }



import heapq
import logging
from typing import Any, Protocol, runtime_checkable

import numpy as np
import torch
import torch.distributed
from lightning_fabric.utilities.distributed import _DatasetSamplerWrapper
from torch.utils.data import BatchSampler, DistributedSampler
from typing_extensions import override

log = logging.getLogger(__name__)


def _all_gather(tensor: torch.Tensor, device: torch.device | None = None):
    gathered = [
        torch.zeros_like(tensor, device=device)
        for _ in range(torch.distributed.get_world_size())
    ]
    _ = torch.distributed.all_gather(gathered, tensor)
    return gathered


# @numba.njit
def _balanced_partition(sizes: np.ndarray, num_parts: int):
    """
    Greedily partition the given set by always inserting
    the largest element into the smallest partition.
    """
    sort_idx = np.argsort(-sizes)  # Sort in descending order
    heap = []
    for idx in sort_idx[:num_parts]:
        heap.append((sizes[idx], [idx]))
    heapq.heapify(heap)
    for idx in sort_idx[num_parts:]:
        smallest_part = heapq.heappop(heap)
        new_size = smallest_part[0] + sizes[idx]
        new_idx = smallest_part[1] + [idx]
        heapq.heappush(heap, (new_size, new_idx))
    idx_balanced = [part[1] for part in heap]
    return idx_balanced


@runtime_checkable
class DatasetWithSizes(Protocol):
    def data_sizes(self, indices: list[int]) -> np.ndarray: ...


@runtime_checkable
class DataSizesFunction(Protocol):
    def __call__(self, dataset: Any, indices: list[int]) -> np.ndarray: ...


class BalancedBatchSampler(BatchSampler):
    @staticmethod
    def _unwrap_dataset(dataset: Any):
        # Lightning's DistributedSampler wraps the dataset in a _DatasetSamplerWrapper,
        # so we need to unwrap it to get the actual dataset.
        if isinstance(dataset, _DatasetSamplerWrapper):
            if (data_source := getattr(dataset._sampler, "data_source", None)) is None:
                raise ValueError("Could not unwrap dataset from _DatasetSamplerWrapper")
            return data_source
        return dataset

    @property
    def distributed_sampler(self):
        if not isinstance(self.sampler, DistributedSampler):
            raise ValueError(
                f"Sampler must be a DistributedSampler, got {type(self.sampler)}"
            )
        return self.sampler

    def __init__(
        self,
        sampler: DistributedSampler,
        *,
        batch_size: int,
        device: torch.device,
        drop_last: bool = False,
        data_sizes_fn: DataSizesFunction | None = None,
    ):
        super().__init__(sampler, batch_size, drop_last=drop_last)

        # Validate the dataset
        dataset = self._unwrap_dataset(self.distributed_sampler.dataset)
        # Dataset much either implement `data_sizes`, or we need to provide a custom
        # implementation of the dataset sizes function.
        if isinstance(dataset, DatasetWithSizes):
            log.critical(f"BalancedBatchSampler: Resolved dataset to {type(dataset)}")

        elif self._data_sizes_fn is not None:
            log.critical("BalancedBatchSampler: Using custom data_sizes_fn")
        else:
            raise ValueError(
                "Dataset must implement the `data_sizes` method, "
                "or a custom data_sizes_fn must be provided "
                "to the BalancedBatchSampler."
            )

        self._device = device
        self._data_sizes_fn = data_sizes_fn

        log.info(
            f"Created BalancedBatchSampler with {sampler=}, {batch_size=}, {drop_last=}"
        )

    @staticmethod
    def _dist_enabled():
        return torch.distributed.is_available() and torch.distributed.is_initialized()

    def _dataset_sizes(self, indices: list[int]) -> np.ndarray:
        dataset = self._unwrap_dataset(self.distributed_sampler.dataset)
        # Dataset much either implement `data_sizes`, or we need to provide a custom
        # implementation of the dataset sizes function.
        if isinstance(dataset, DatasetWithSizes):
            return dataset.data_sizes(indices)

        if (data_sizes_fn := self._data_sizes_fn) is not None:
            return data_sizes_fn(dataset, indices)

        raise ValueError(
            "Dataset must implement the `data_sizes` method, "
            "or a custom data_sizes_fn must be provided "
            "to the BalancedBatchSampler."
        )

    @override
    def __iter__(self):
        if not self._dist_enabled():
            yield from super().__iter__()
            return

        for batch_idxs in super().__iter__():
            sizes = self._dataset_sizes(batch_idxs)
            idx_sizes = torch.stack(
                [
                    torch.tensor(batch_idxs, device=self._device),
                    torch.tensor(sizes, device=self._device),
                ]
            )
            idx_sizes_all = _all_gather(idx_sizes, device=self._device)
            idx_sizes_all = torch.cat(idx_sizes_all, dim=-1).cpu()
            idx_all = idx_sizes_all[0]
            sizes_all = idx_sizes_all[1]

            local_idx_balanced = _balanced_partition(
                sizes_all.numpy(), num_parts=self.distributed_sampler.num_replicas
            )
            # Since DistributedSampler pads the last batch
            # this should always have an entry for each replica.
            yield idx_all[local_idx_balanced[self.distributed_sampler.rank]].tolist()



import logging
from typing import Literal

from typing_extensions import override

from ._base import BaseProfilerConfig

log = logging.getLogger(__name__)


class SimpleProfilerConfig(BaseProfilerConfig):
    name: Literal["simple"] = "simple"

    extended: bool = True
    """
    If ``True``, adds extra columns representing number of calls and percentage of
        total time spent onrespective action.
    """

    @override
    def create_profiler(self, root_config):
        from lightning.pytorch.profilers.simple import SimpleProfiler

        if (dirpath := self.dirpath) is None:
            dirpath = root_config.directory.resolve_subdirectory(
                root_config.id, "profile"
            )

        if (filename := self.filename) is None:
            filename = f"{root_config.id}_profile.txt"

        return SimpleProfiler(
            extended=self.extended,
            dirpath=dirpath,
            filename=filename,
        )



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.loggers.wandb import BaseLoggerConfig as BaseLoggerConfig
from nshtrainer.loggers.wandb import WandbLoggerConfig as WandbLoggerConfig
from nshtrainer.loggers.wandb import WandbUploadCodeConfig as WandbUploadCodeConfig
from nshtrainer.loggers.wandb import CallbackConfigBase as CallbackConfigBase
from nshtrainer.loggers.wandb import WandbWatchConfig as WandbWatchConfig

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.lr_scheduler.reduce_lr_on_plateau import ReduceLROnPlateauConfig as ReduceLROnPlateauConfig
from nshtrainer.lr_scheduler.reduce_lr_on_plateau import LRSchedulerConfigBase as LRSchedulerConfigBase
from nshtrainer.lr_scheduler.reduce_lr_on_plateau import MetricConfig as MetricConfig

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.optimizer import OptimizerConfigBase as OptimizerConfigBase
from nshtrainer.optimizer import AdamWConfig as AdamWConfig

# Type aliases
from nshtrainer.optimizer import OptimizerConfig as OptimizerConfig

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.model import MetricConfig as MetricConfig
from nshtrainer.model import BaseConfig as BaseConfig
from nshtrainer.model import DirectoryConfig as DirectoryConfig
from nshtrainer.model import TrainerConfig as TrainerConfig
from nshtrainer.model.base import EnvironmentConfig as EnvironmentConfig
from nshtrainer.model.config import CallbackConfigBase as CallbackConfigBase

# Type aliases

# Submodule exports
from . import base as base
from . import config as config
from . import mixins as mixins



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.profiler.pytorch import PyTorchProfilerConfig as PyTorchProfilerConfig
from nshtrainer.profiler.pytorch import BaseProfilerConfig as BaseProfilerConfig

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.callbacks.checkpoint import LastCheckpointCallbackConfig as LastCheckpointCallbackConfig
from nshtrainer.callbacks.checkpoint._base import CheckpointMetadata as CheckpointMetadata
from nshtrainer.callbacks.checkpoint._base import BaseCheckpointCallbackConfig as BaseCheckpointCallbackConfig
from nshtrainer.callbacks.checkpoint._base import CallbackConfigBase as CallbackConfigBase
from nshtrainer.callbacks.checkpoint import OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig
from nshtrainer.callbacks.checkpoint import BestCheckpointCallbackConfig as BestCheckpointCallbackConfig
from nshtrainer.callbacks.checkpoint.best_checkpoint import MetricConfig as MetricConfig

# Type aliases

# Submodule exports
from . import _base as _base
from . import best_checkpoint as best_checkpoint
from . import last_checkpoint as last_checkpoint
from . import on_exception_checkpoint as on_exception_checkpoint



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.callbacks.debug_flag import DebugFlagCallbackConfig as DebugFlagCallbackConfig
from nshtrainer.callbacks.debug_flag import CallbackConfigBase as CallbackConfigBase

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.callbacks.checkpoint.best_checkpoint import BestCheckpointCallbackConfig as BestCheckpointCallbackConfig
from nshtrainer.callbacks.checkpoint.best_checkpoint import CheckpointMetadata as CheckpointMetadata
from nshtrainer.callbacks.checkpoint.best_checkpoint import BaseCheckpointCallbackConfig as BaseCheckpointCallbackConfig
from nshtrainer.callbacks.checkpoint.best_checkpoint import MetricConfig as MetricConfig

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.loggers import TensorboardLoggerConfig as TensorboardLoggerConfig
from nshtrainer.loggers import BaseLoggerConfig as BaseLoggerConfig
from nshtrainer.loggers import WandbLoggerConfig as WandbLoggerConfig
from nshtrainer.loggers.wandb import WandbUploadCodeConfig as WandbUploadCodeConfig
from nshtrainer.loggers.wandb import CallbackConfigBase as CallbackConfigBase
from nshtrainer.loggers.wandb import WandbWatchConfig as WandbWatchConfig
from nshtrainer.loggers import CSVLoggerConfig as CSVLoggerConfig

# Type aliases
from nshtrainer.loggers import LoggerConfig as LoggerConfig

# Submodule exports
from . import _base as _base
from . import csv as csv
from . import tensorboard as tensorboard
from . import wandb as wandb



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.profiler.simple import BaseProfilerConfig as BaseProfilerConfig
from nshtrainer.profiler.simple import SimpleProfilerConfig as SimpleProfilerConfig

# Type aliases

# Submodule exports



class SlurmParseException(Exception):
    pass


def _expand_id_suffix(suffix_parts: str) -> list[str]:
    """Parse the a suffix formatted like "1-3,5,8" into
    the list of numeric values 1,2,3,5,8.
    """
    suffixes = []
    for suffix_part in suffix_parts.split(","):
        if "-" in suffix_part:
            low, high = suffix_part.split("-")
            int_length = len(low)
            for num in range(int(low), int(high) + 1):
                suffixes.append(f"{num:0{int_length}}")
        else:
            suffixes.append(suffix_part)
    return suffixes


def _parse_node_group(node_list: str, pos: int, parsed: list[str]) -> int:
    """Parse a node group of the form PREFIX[1-3,5,8] and return
    the position in the string at which the parsing stopped
    """
    prefixes = [""]
    while pos < len(node_list):
        c = node_list[pos]
        if c == ",":
            parsed.extend(prefixes)
            return pos + 1
        if c == "[":
            last_pos = node_list.index("]", pos)
            suffixes = _expand_id_suffix(node_list[pos + 1 : last_pos])
            prefixes = [prefix + suffix for prefix in prefixes for suffix in suffixes]
            pos = last_pos + 1
        else:
            for i, prefix in enumerate(prefixes):
                prefixes[i] = prefix + c
            pos += 1
    parsed.extend(prefixes)
    return pos


def parse_slurm_node_list(node_list: str):
    pos = 0
    parsed: list[str] = []
    while pos < len(node_list):
        pos = _parse_node_group(node_list, pos, parsed)
    return parsed



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.profiler.advanced import AdvancedProfilerConfig as AdvancedProfilerConfig
from nshtrainer.profiler.advanced import BaseProfilerConfig as BaseProfilerConfig

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer._checkpoint.loader import UserProvidedPathCheckpointStrategyConfig as UserProvidedPathCheckpointStrategyConfig
from nshtrainer._checkpoint.loader import MetricConfig as MetricConfig
from nshtrainer._checkpoint.loader import CheckpointLoadingConfig as CheckpointLoadingConfig
from nshtrainer._checkpoint.loader import CheckpointMetadata as CheckpointMetadata
from nshtrainer._checkpoint.loader import BestCheckpointStrategyConfig as BestCheckpointStrategyConfig
from nshtrainer._checkpoint.loader import LastCheckpointStrategyConfig as LastCheckpointStrategyConfig

# Type aliases
from nshtrainer._checkpoint.loader import CheckpointLoadingStrategyConfig as CheckpointLoadingStrategyConfig

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.nn.nonlinearity import TanhNonlinearityConfig as TanhNonlinearityConfig
from nshtrainer.nn.nonlinearity import ReLUNonlinearityConfig as ReLUNonlinearityConfig
from nshtrainer.nn.nonlinearity import PReLUConfig as PReLUConfig
from nshtrainer.nn.nonlinearity import SiLUNonlinearityConfig as SiLUNonlinearityConfig
from nshtrainer.nn.nonlinearity import LeakyReLUNonlinearityConfig as LeakyReLUNonlinearityConfig
from nshtrainer.nn.nonlinearity import SoftsignNonlinearityConfig as SoftsignNonlinearityConfig
from nshtrainer.nn.nonlinearity import MishNonlinearityConfig as MishNonlinearityConfig
from nshtrainer.nn.nonlinearity import SigmoidNonlinearityConfig as SigmoidNonlinearityConfig
from nshtrainer.nn.nonlinearity import SoftplusNonlinearityConfig as SoftplusNonlinearityConfig
from nshtrainer.nn.nonlinearity import SwiGLUNonlinearityConfig as SwiGLUNonlinearityConfig
from nshtrainer.nn.nonlinearity import ELUNonlinearityConfig as ELUNonlinearityConfig
from nshtrainer.nn.nonlinearity import BaseNonlinearityConfig as BaseNonlinearityConfig
from nshtrainer.nn.nonlinearity import SoftmaxNonlinearityConfig as SoftmaxNonlinearityConfig
from nshtrainer.nn.nonlinearity import GELUNonlinearityConfig as GELUNonlinearityConfig
from nshtrainer.nn.nonlinearity import SwishNonlinearityConfig as SwishNonlinearityConfig

# Type aliases
from nshtrainer.nn.nonlinearity import NonlinearityConfig as NonlinearityConfig

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer._hf_hub import HuggingFaceHubAutoCreateConfig as HuggingFaceHubAutoCreateConfig
from nshtrainer._hf_hub import CallbackConfigBase as CallbackConfigBase
from nshtrainer._hf_hub import HuggingFaceHubConfig as HuggingFaceHubConfig

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.util._environment_info import EnvironmentSnapshotConfig as EnvironmentSnapshotConfig
from nshtrainer.util._environment_info import EnvironmentHardwareConfig as EnvironmentHardwareConfig
from nshtrainer.util._environment_info import EnvironmentPackageConfig as EnvironmentPackageConfig
from nshtrainer.util._environment_info import GitRepositoryConfig as GitRepositoryConfig
from nshtrainer.util._environment_info import EnvironmentGPUConfig as EnvironmentGPUConfig
from nshtrainer.util._environment_info import EnvironmentCUDAConfig as EnvironmentCUDAConfig
from nshtrainer.util._environment_info import EnvironmentLinuxEnvironmentConfig as EnvironmentLinuxEnvironmentConfig
from nshtrainer.util._environment_info import EnvironmentClassInformationConfig as EnvironmentClassInformationConfig
from nshtrainer.util._environment_info import EnvironmentSLURMInformationConfig as EnvironmentSLURMInformationConfig
from nshtrainer.util._environment_info import EnvironmentConfig as EnvironmentConfig
from nshtrainer.util._environment_info import EnvironmentLSFInformationConfig as EnvironmentLSFInformationConfig

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.lr_scheduler import LRSchedulerConfigBase as LRSchedulerConfigBase
from nshtrainer.lr_scheduler import LinearWarmupCosineDecayLRSchedulerConfig as LinearWarmupCosineDecayLRSchedulerConfig
from nshtrainer.lr_scheduler import ReduceLROnPlateauConfig as ReduceLROnPlateauConfig
from nshtrainer.lr_scheduler.reduce_lr_on_plateau import MetricConfig as MetricConfig

# Type aliases
from nshtrainer.lr_scheduler.linear_warmup_cosine import DurationConfig as DurationConfig
from nshtrainer.lr_scheduler import LRSchedulerConfig as LRSchedulerConfig

# Submodule exports
from . import _base as _base
from . import linear_warmup_cosine as linear_warmup_cosine
from . import reduce_lr_on_plateau as reduce_lr_on_plateau



import copy
import logging
from collections.abc import Callable, Iterable, Mapping, Sequence
from pathlib import Path
from typing import Generic

import nshrunner as nr
from nshrunner._submit import screen
from typing_extensions import TypeVar, TypeVarTuple, Unpack, deprecated, override

from .model.config import BaseConfig

TConfig = TypeVar("TConfig", bound=BaseConfig, infer_variance=True)
TArguments = TypeVarTuple("TArguments", default=Unpack[tuple[()]])
TReturn = TypeVar("TReturn", infer_variance=True)


@deprecated("Use nshrunner.Runner instead.")
class Runner(
    nr.Runner[TReturn, TConfig, Unpack[TArguments]],
    Generic[TReturn, TConfig, Unpack[TArguments]],
):
    @override
    def __init__(
        self,
        run_fn: Callable[[TConfig, Unpack[TArguments]], TReturn],
        config: nr.RunnerConfig | None = None,
    ):
        if config is None:
            working_dir = Path.cwd() / "nshrunner"
            working_dir.mkdir(exist_ok=True)

            logging.warning(
                f"`config` is not provided. Using default working directory of {working_dir}."
            )
            config = nr.RunnerConfig(working_dir=working_dir)

        super().__init__(run_fn, config)

    def fast_dev_run(
        self,
        runs: Iterable[tuple[TConfig, Unpack[TArguments]]],
        n_batches: int = 1,
        *,
        env: Mapping[str, str] | None = None,
    ):
        runs_updated: list[tuple[TConfig, Unpack[TArguments]]] = []
        for args in runs:
            config = copy.deepcopy(args[0])
            config.trainer.fast_dev_run = n_batches
            runs_updated.append((config, *args[1:]))
        del runs

        return self.local(runs_updated, env=env)

    def fast_dev_run_generator(
        self,
        runs: Iterable[tuple[TConfig, Unpack[TArguments]]],
        n_batches: int = 1,
        *,
        env: Mapping[str, str] | None = None,
    ):
        runs_updated: list[tuple[TConfig, Unpack[TArguments]]] = []
        for args in runs:
            config = copy.deepcopy(args[0])
            config.trainer.fast_dev_run = n_batches
            runs_updated.append((config, *args[1:]))
        del runs

        return self.local_generator(runs_updated, env=env)

    def fast_dev_run_session(
        self,
        runs: Iterable[tuple[TConfig, Unpack[TArguments]]],
        options: screen.ScreenJobKwargs = {},
        n_batches: int = 1,
        *,
        snapshot: nr.Snapshot,
        setup_commands: Sequence[str] | None = None,
        env: Mapping[str, str] | None = None,
        activate_venv: bool = True,
        print_command: bool = True,
    ):
        runs_updated: list[tuple[TConfig, Unpack[TArguments]]] = []
        for args in runs:
            config = copy.deepcopy(args[0])
            config.trainer.fast_dev_run = n_batches
            runs_updated.append((config, *args[1:]))
        del runs

        return self.session(
            runs_updated,
            options,
            snapshot=snapshot,
            setup_commands=setup_commands,
            env=env,
            activate_venv=activate_venv,
            print_command=print_command,
        )



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.util.config import EpochsConfig as EpochsConfig
from nshtrainer.util.config import StepsConfig as StepsConfig
from nshtrainer.util.config import DTypeConfig as DTypeConfig

# Type aliases
from nshtrainer.util.config import DurationConfig as DurationConfig

# Submodule exports
from . import dtype as dtype
from . import duration as duration



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.lr_scheduler.linear_warmup_cosine import LinearWarmupCosineDecayLRSchedulerConfig as LinearWarmupCosineDecayLRSchedulerConfig
from nshtrainer.lr_scheduler.linear_warmup_cosine import LRSchedulerConfigBase as LRSchedulerConfigBase

# Type aliases
from nshtrainer.lr_scheduler.linear_warmup_cosine import DurationConfig as DurationConfig

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.profiler import BaseProfilerConfig as BaseProfilerConfig
from nshtrainer.profiler import PyTorchProfilerConfig as PyTorchProfilerConfig
from nshtrainer.profiler import AdvancedProfilerConfig as AdvancedProfilerConfig
from nshtrainer.profiler import SimpleProfilerConfig as SimpleProfilerConfig

# Type aliases
from nshtrainer.profiler import ProfilerConfig as ProfilerConfig

# Submodule exports
from . import _base as _base
from . import advanced as advanced
from . import pytorch as pytorch
from . import simple as simple



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.callbacks.early_stopping import EarlyStoppingConfig as EarlyStoppingConfig
from nshtrainer.callbacks.early_stopping import CallbackConfigBase as CallbackConfigBase
from nshtrainer.callbacks.early_stopping import MetricConfig as MetricConfig

# Type aliases

# Submodule exports



import logging
import math
from typing import Literal

from lightning.fabric.utilities.rank_zero import _get_rank
from lightning.pytorch import Trainer
from lightning.pytorch.callbacks import EarlyStopping as _EarlyStopping
from lightning.pytorch.utilities.rank_zero import rank_prefixed_message
from typing_extensions import override

from ..metrics._config import MetricConfig
from .base import CallbackConfigBase

log = logging.getLogger(__name__)


class EarlyStoppingConfig(CallbackConfigBase):
    name: Literal["early_stopping"] = "early_stopping"

    metric: MetricConfig | None = None
    """
    The metric to monitor for early stopping.
    If None, the primary metric will be used.
    """

    patience: int
    """
    Number of epochs with no improvement after which training will be stopped.
    """

    min_delta: float = 1.0e-8
    """
    Minimum change in the monitored quantity to qualify as an improvement.
    """

    min_lr: float | None = None
    """
    Minimum learning rate. If the learning rate of the model is less than this value,
    the training will be stopped.
    """

    strict: bool = True
    """
    Whether to enforce that the monitored quantity must improve by at least `min_delta`
    to qualify as an improvement.
    """

    @override
    def create_callbacks(self, root_config):
        if (metric := self.metric) is None and (
            metric := root_config.primary_metric
        ) is None:
            raise ValueError(
                "Either `metric` or `root_config.primary_metric` must be set to use EarlyStopping."
            )

        yield EarlyStopping(self, metric)


class EarlyStopping(_EarlyStopping):
    def __init__(self, config: EarlyStoppingConfig, metric: MetricConfig):
        self.config = config
        self.metric = metric
        del config, metric

        super().__init__(
            monitor=self.metric.validation_monitor,
            mode=self.metric.mode,
            patience=self.config.patience,
            min_delta=self.config.min_delta,
            strict=self.config.strict,
        )

    @override
    @staticmethod
    def _log_info(trainer: Trainer | None, message: str, log_rank_zero_only: bool):
        rank = _get_rank()
        if trainer is not None and trainer.world_size <= 1:
            rank = None
        message = rank_prefixed_message(message, rank)
        if rank is None or not log_rank_zero_only or rank == 0:
            log.critical(message)

    @override
    def _run_early_stopping_check(self, trainer: Trainer):
        """Checks whether the early stopping condition is met and if so tells the trainer to stop the training."""
        logs = trainer.callback_metrics

        # Disable early_stopping with fast_dev_run
        if getattr(trainer, "fast_dev_run", False):
            return

        should_stop, reason = False, None

        if not should_stop:
            should_stop, reason = self._evaluate_stopping_criteria_min_lr(trainer)

        # If metric present
        if not should_stop and self._validate_condition_metric(logs):
            current = logs[self.monitor].squeeze()
            should_stop, reason = self._evaluate_stopping_criteria(current)

        # stop every ddp process if any world process decides to stop
        should_stop = trainer.strategy.reduce_boolean_decision(should_stop, all=False)
        trainer.should_stop = trainer.should_stop or should_stop
        if should_stop:
            self.stopped_epoch = trainer.current_epoch
        if reason and self.verbose:
            self._log_info(trainer, reason, self.log_rank_zero_only)

    def _evaluate_stopping_criteria_min_lr(
        self, trainer: Trainer
    ) -> tuple[bool, str | None]:
        if self.config.min_lr is None:
            return False, None

        # Get the maximum LR across all param groups in all optimizers
        model_max_lr = max(
            [
                param_group["lr"]
                for optimizer in trainer.optimizers
                for param_group in optimizer.param_groups
            ]
        )
        if not isinstance(model_max_lr, float) or not math.isfinite(model_max_lr):
            return False, None

        # If the maximum LR is less than the minimum LR, stop training
        if model_max_lr >= self.config.min_lr:
            return False, None

        return True, (
            "Stopping threshold reached: "
            f"The maximum LR of the model across all param groups is {model_max_lr:.2e} "
            f"which is less than the minimum LR {self.config.min_lr:.2e}"
        )

    def on_early_stopping(self, trainer: Trainer):
        pass



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.model.config import MetricConfig as MetricConfig
from nshtrainer.model.config import BaseConfig as BaseConfig
from nshtrainer.model.config import DirectoryConfig as DirectoryConfig
from nshtrainer.model.config import TrainerConfig as TrainerConfig
from nshtrainer.model.config import EnvironmentConfig as EnvironmentConfig
from nshtrainer.model.config import CallbackConfigBase as CallbackConfigBase

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer import MetricConfig as MetricConfig
from nshtrainer import BaseConfig as BaseConfig
from nshtrainer._hf_hub import HuggingFaceHubAutoCreateConfig as HuggingFaceHubAutoCreateConfig
from nshtrainer._hf_hub import CallbackConfigBase as CallbackConfigBase
from nshtrainer._hf_hub import HuggingFaceHubConfig as HuggingFaceHubConfig
from nshtrainer.optimizer import OptimizerConfigBase as OptimizerConfigBase
from nshtrainer.optimizer import AdamWConfig as AdamWConfig
from nshtrainer.model import DirectoryConfig as DirectoryConfig
from nshtrainer.callbacks import DirectorySetupConfig as DirectorySetupConfig
from nshtrainer.model import TrainerConfig as TrainerConfig
from nshtrainer.model.base import EnvironmentConfig as EnvironmentConfig
from nshtrainer.nn import MLPConfig as MLPConfig
from nshtrainer.nn import BaseNonlinearityConfig as BaseNonlinearityConfig
from nshtrainer.nn import TanhNonlinearityConfig as TanhNonlinearityConfig
from nshtrainer.nn import ReLUNonlinearityConfig as ReLUNonlinearityConfig
from nshtrainer.nn import PReLUConfig as PReLUConfig
from nshtrainer.nn import SiLUNonlinearityConfig as SiLUNonlinearityConfig
from nshtrainer.nn import LeakyReLUNonlinearityConfig as LeakyReLUNonlinearityConfig
from nshtrainer.nn import SoftsignNonlinearityConfig as SoftsignNonlinearityConfig
from nshtrainer.nn import MishNonlinearityConfig as MishNonlinearityConfig
from nshtrainer.nn import SigmoidNonlinearityConfig as SigmoidNonlinearityConfig
from nshtrainer.nn import SoftplusNonlinearityConfig as SoftplusNonlinearityConfig
from nshtrainer.nn.nonlinearity import SwiGLUNonlinearityConfig as SwiGLUNonlinearityConfig
from nshtrainer.nn import ELUNonlinearityConfig as ELUNonlinearityConfig
from nshtrainer.nn import SoftmaxNonlinearityConfig as SoftmaxNonlinearityConfig
from nshtrainer.nn import GELUNonlinearityConfig as GELUNonlinearityConfig
from nshtrainer.nn import SwishNonlinearityConfig as SwishNonlinearityConfig
from nshtrainer.lr_scheduler import LRSchedulerConfigBase as LRSchedulerConfigBase
from nshtrainer.lr_scheduler import LinearWarmupCosineDecayLRSchedulerConfig as LinearWarmupCosineDecayLRSchedulerConfig
from nshtrainer.lr_scheduler import ReduceLROnPlateauConfig as ReduceLROnPlateauConfig
from nshtrainer.loggers import TensorboardLoggerConfig as TensorboardLoggerConfig
from nshtrainer.loggers import BaseLoggerConfig as BaseLoggerConfig
from nshtrainer.loggers import WandbLoggerConfig as WandbLoggerConfig
from nshtrainer.callbacks import WandbUploadCodeConfig as WandbUploadCodeConfig
from nshtrainer.callbacks import WandbWatchConfig as WandbWatchConfig
from nshtrainer.loggers import CSVLoggerConfig as CSVLoggerConfig
from nshtrainer.util._environment_info import EnvironmentSnapshotConfig as EnvironmentSnapshotConfig
from nshtrainer.util._environment_info import EnvironmentHardwareConfig as EnvironmentHardwareConfig
from nshtrainer.util._environment_info import EnvironmentPackageConfig as EnvironmentPackageConfig
from nshtrainer.util._environment_info import GitRepositoryConfig as GitRepositoryConfig
from nshtrainer.util._environment_info import EnvironmentGPUConfig as EnvironmentGPUConfig
from nshtrainer.util._environment_info import EnvironmentCUDAConfig as EnvironmentCUDAConfig
from nshtrainer.util._environment_info import EnvironmentLinuxEnvironmentConfig as EnvironmentLinuxEnvironmentConfig
from nshtrainer.util._environment_info import EnvironmentClassInformationConfig as EnvironmentClassInformationConfig
from nshtrainer.util._environment_info import EnvironmentSLURMInformationConfig as EnvironmentSLURMInformationConfig
from nshtrainer.util._environment_info import EnvironmentLSFInformationConfig as EnvironmentLSFInformationConfig
from nshtrainer.util.config import EpochsConfig as EpochsConfig
from nshtrainer.util.config import StepsConfig as StepsConfig
from nshtrainer.util.config import DTypeConfig as DTypeConfig
from nshtrainer.trainer._config import CheckpointLoadingConfig as CheckpointLoadingConfig
from nshtrainer.callbacks import BestCheckpointCallbackConfig as BestCheckpointCallbackConfig
from nshtrainer.trainer._config import CheckpointSavingConfig as CheckpointSavingConfig
from nshtrainer.callbacks import SharedParametersConfig as SharedParametersConfig
from nshtrainer.trainer._config import SanityCheckingConfig as SanityCheckingConfig
from nshtrainer.trainer._config import ReproducibilityConfig as ReproducibilityConfig
from nshtrainer.trainer._config import LoggingConfig as LoggingConfig
from nshtrainer.callbacks import LastCheckpointCallbackConfig as LastCheckpointCallbackConfig
from nshtrainer.callbacks import OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig
from nshtrainer.callbacks import RLPSanityChecksConfig as RLPSanityChecksConfig
from nshtrainer.callbacks import EarlyStoppingConfig as EarlyStoppingConfig
from nshtrainer.trainer._config import OptimizationConfig as OptimizationConfig
from nshtrainer.callbacks import DebugFlagCallbackConfig as DebugFlagCallbackConfig
from nshtrainer.trainer._config import GradientClippingConfig as GradientClippingConfig
from nshtrainer._checkpoint.loader import CheckpointMetadata as CheckpointMetadata
from nshtrainer._checkpoint.loader import UserProvidedPathCheckpointStrategyConfig as UserProvidedPathCheckpointStrategyConfig
from nshtrainer._checkpoint.loader import BestCheckpointStrategyConfig as BestCheckpointStrategyConfig
from nshtrainer._checkpoint.loader import LastCheckpointStrategyConfig as LastCheckpointStrategyConfig
from nshtrainer.callbacks import PrintTableMetricsConfig as PrintTableMetricsConfig
from nshtrainer.callbacks import ThroughputMonitorConfig as ThroughputMonitorConfig
from nshtrainer.callbacks import GradientSkippingConfig as GradientSkippingConfig
from nshtrainer.callbacks import EMAConfig as EMAConfig
from nshtrainer.callbacks.actsave import ActSaveConfig as ActSaveConfig
from nshtrainer.callbacks import FiniteChecksConfig as FiniteChecksConfig
from nshtrainer.callbacks import NormLoggingConfig as NormLoggingConfig
from nshtrainer.callbacks import EpochTimerConfig as EpochTimerConfig
from nshtrainer.callbacks.checkpoint._base import BaseCheckpointCallbackConfig as BaseCheckpointCallbackConfig
from nshtrainer.profiler import BaseProfilerConfig as BaseProfilerConfig
from nshtrainer.profiler import PyTorchProfilerConfig as PyTorchProfilerConfig
from nshtrainer.profiler import AdvancedProfilerConfig as AdvancedProfilerConfig
from nshtrainer.profiler import SimpleProfilerConfig as SimpleProfilerConfig

# Type aliases
from nshtrainer.optimizer import OptimizerConfig as OptimizerConfig
from nshtrainer.loggers import LoggerConfig as LoggerConfig
from nshtrainer.nn import NonlinearityConfig as NonlinearityConfig
from nshtrainer.util.config import DurationConfig as DurationConfig
from nshtrainer.lr_scheduler import LRSchedulerConfig as LRSchedulerConfig
from nshtrainer.callbacks import CallbackConfig as CallbackConfig
from nshtrainer.trainer._config import CheckpointCallbackConfig as CheckpointCallbackConfig
from nshtrainer.profiler import ProfilerConfig as ProfilerConfig
from nshtrainer._checkpoint.loader import CheckpointLoadingStrategyConfig as CheckpointLoadingStrategyConfig

# Submodule exports
from . import _checkpoint as _checkpoint
from . import _directory as _directory
from . import _hf_hub as _hf_hub
from . import callbacks as callbacks
from . import loggers as loggers
from . import lr_scheduler as lr_scheduler
from . import metrics as metrics
from . import model as model
from . import nn as nn
from . import optimizer as optimizer
from . import profiler as profiler
from . import runner as runner
from . import trainer as trainer
from . import util as util



import contextlib
import datetime
import logging
import os
from pathlib import Path
from typing import Any, Literal

from lightning.pytorch import Trainer as LightningTrainer
from lightning.pytorch.callbacks import OnExceptionCheckpoint as _OnExceptionCheckpoint
from typing_extensions import override

from ..base import CallbackConfigBase

log = logging.getLogger(__name__)


@contextlib.contextmanager
def _monkey_patch_disable_barrier(trainer: LightningTrainer):
    """
    Monkey-patch the strategy instance to make the barrier operation a no-op.

    We do this because `save_checkpoint` calls `barrier`. This is okay in most
    cases, but when we want to save a checkpoint in the case of an exception,
    `barrier` causes a deadlock. So we monkey-patch the strategy instance to
    make the barrier operation a no-op.
    """

    # We monkey-patch the barrier method to do nothing.
    original_barrier = trainer.strategy.barrier

    def new_barrier(*args, **kwargs):
        log.warning("Monkey-patched no-op barrier.")
        pass

    trainer.strategy.barrier = new_barrier
    log.warning("Monkey-patched barrier to no-op.")

    try:
        yield
    finally:
        trainer.strategy.barrier = original_barrier
        log.warning("Reverted monkey-patched barrier.")


class OnExceptionCheckpointCallbackConfig(CallbackConfigBase):
    name: Literal["on_exception_checkpoint"] = "on_exception_checkpoint"

    dirpath: str | Path | None = None
    """Directory path to save the checkpoint file."""

    filename: str | None = None
    """Checkpoint filename. This must not include the extension. If `None`, `on_exception_{id}_{timestamp}` is used."""

    @override
    def create_callbacks(self, root_config):
        dirpath = self.dirpath or root_config.directory.resolve_subdirectory(
            root_config.id, "checkpoint"
        )

        if not (filename := self.filename):
            filename = f"on_exception_{root_config.id}"
        yield OnExceptionCheckpoint(self, dirpath=Path(dirpath), filename=filename)


class OnExceptionCheckpoint(_OnExceptionCheckpoint):
    @override
    def __init__(
        self,
        config: OnExceptionCheckpointCallbackConfig,
        dirpath: Path,
        filename: str,
    ):
        self.config = config
        del config

        dirpath = dirpath / "on_exception"
        dirpath.mkdir(parents=True, exist_ok=True)

        super().__init__(dirpath, filename)

    @property
    @override
    def ckpt_path(self) -> str:
        ckpt_path = super().ckpt_path

        # Remve the extension and add the current timestamp
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        ckpt_path, ext = os.path.splitext(ckpt_path)
        return f"{ckpt_path}_{timestamp}{ext}"

    @override
    def on_exception(self, trainer: LightningTrainer, *args: Any, **kwargs: Any):
        # Monkey-patch the strategy instance to make the barrier operation a no-op.
        # We do this because `save_checkpoint` calls `barrier`. This is okay in most
        # cases, but when we want to save a checkpoint in the case of an exception,
        # `barrier` causes a deadlock. So we monkey-patch the strategy instance to
        # make the barrier operation a no-op.
        with _monkey_patch_disable_barrier(trainer):
            return super().on_exception(trainer, *args, **kwargs)



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer._checkpoint.metadata import EnvironmentConfig as EnvironmentConfig
from nshtrainer._checkpoint.metadata import CheckpointMetadata as CheckpointMetadata

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer._directory import DirectoryConfig as DirectoryConfig
from nshtrainer._directory import DirectorySetupConfig as DirectorySetupConfig

# Type aliases
from nshtrainer._directory import LoggerConfig as LoggerConfig

# Submodule exports



import copy
import fnmatch
import importlib.util
import logging
from typing import Literal

import torch
from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks import Callback
from typing_extensions import override

from .base import CallbackConfigBase

log = logging.getLogger(__name__)


class PrintTableMetricsCallback(Callback):
    """Prints a table with the metrics in columns on every epoch end."""

    def __init__(
        self,
        metric_patterns: list[str] | None = None,
    ) -> None:
        self.metrics: list = []
        self.rich_available = importlib.util.find_spec("rich") is not None
        self.metric_patterns = metric_patterns

        if not self.rich_available:
            log.warning(
                "rich is not installed. Please install it to use PrintTableMetricsCallback."
            )

    @override
    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        if not self.rich_available:
            return

        metrics_dict = copy.copy(trainer.callback_metrics)
        # Filter metrics based on the patterns
        if self.metric_patterns is not None:
            metrics_dict = {
                key: value
                for key, value in metrics_dict.items()
                if any(
                    fnmatch.fnmatch(key, pattern) for pattern in self.metric_patterns
                )
            }
        self.metrics.append(metrics_dict)

        from rich.console import Console

        console = Console()
        table = self.create_metrics_table()
        console.print(table)

    def create_metrics_table(self):
        from rich.table import Table

        table = Table(show_header=True, header_style="bold magenta")

        # Add columns to the table based on the keys in the first metrics dictionary
        for key in self.metrics[0].keys():
            table.add_column(key)

        # Add rows to the table based on the metrics dictionaries
        for metric_dict in self.metrics:
            values: list[str] = []
            for value in metric_dict.values():
                if torch.is_tensor(value):
                    value = float(value.item())
                values.append(str(value))
            table.add_row(*values)

        return table


class PrintTableMetricsConfig(CallbackConfigBase):
    """Configuration class for PrintTableMetricsCallback."""

    name: Literal["print_table_metrics"] = "print_table_metrics"

    enabled: bool = True
    """Whether to enable the callback or not."""

    metric_patterns: list[str] | None = None
    """List of patterns to filter the metrics to be displayed. If None, all metrics are displayed."""

    @override
    def create_callbacks(self, root_config):
        yield PrintTableMetricsCallback(metric_patterns=self.metric_patterns)



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.util.config.duration import EpochsConfig as EpochsConfig
from nshtrainer.util.config.duration import StepsConfig as StepsConfig

# Type aliases
from nshtrainer.util.config.duration import DurationConfig as DurationConfig

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.callbacks import PrintTableMetricsConfig as PrintTableMetricsConfig
from nshtrainer.callbacks import CallbackConfigBase as CallbackConfigBase
from nshtrainer.callbacks import DebugFlagCallbackConfig as DebugFlagCallbackConfig
from nshtrainer.callbacks import ThroughputMonitorConfig as ThroughputMonitorConfig
from nshtrainer.callbacks import GradientSkippingConfig as GradientSkippingConfig
from nshtrainer.callbacks import RLPSanityChecksConfig as RLPSanityChecksConfig
from nshtrainer.callbacks import WandbUploadCodeConfig as WandbUploadCodeConfig
from nshtrainer.callbacks import EarlyStoppingConfig as EarlyStoppingConfig
from nshtrainer.callbacks.early_stopping import MetricConfig as MetricConfig
from nshtrainer.callbacks import WandbWatchConfig as WandbWatchConfig
from nshtrainer.callbacks import EMAConfig as EMAConfig
from nshtrainer.callbacks import DirectorySetupConfig as DirectorySetupConfig
from nshtrainer.callbacks.actsave import ActSaveConfig as ActSaveConfig
from nshtrainer.callbacks import FiniteChecksConfig as FiniteChecksConfig
from nshtrainer.callbacks import NormLoggingConfig as NormLoggingConfig
from nshtrainer.callbacks import BestCheckpointCallbackConfig as BestCheckpointCallbackConfig
from nshtrainer.callbacks import OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig
from nshtrainer.callbacks import SharedParametersConfig as SharedParametersConfig
from nshtrainer.callbacks import EpochTimerConfig as EpochTimerConfig
from nshtrainer.callbacks import LastCheckpointCallbackConfig as LastCheckpointCallbackConfig
from nshtrainer.callbacks.checkpoint._base import CheckpointMetadata as CheckpointMetadata
from nshtrainer.callbacks.checkpoint._base import BaseCheckpointCallbackConfig as BaseCheckpointCallbackConfig

# Type aliases
from nshtrainer.callbacks import CallbackConfig as CallbackConfig

# Submodule exports
from . import actsave as actsave
from . import base as base
from . import checkpoint as checkpoint
from . import debug_flag as debug_flag
from . import directory_setup as directory_setup
from . import early_stopping as early_stopping
from . import ema as ema
from . import finite_checks as finite_checks
from . import gradient_skipping as gradient_skipping
from . import norm_logging as norm_logging
from . import print_table as print_table
from . import rlp_sanity_checks as rlp_sanity_checks
from . import shared_parameters as shared_parameters
from . import throughput_monitor as throughput_monitor
from . import timer as timer
from . import wandb_upload_code as wandb_upload_code
from . import wandb_watch as wandb_watch



from typing import Annotated, TypeAlias

import nshconfig as C

from ._base import BaseLoggerConfig as BaseLoggerConfig
from .csv import CSVLoggerConfig as CSVLoggerConfig
from .tensorboard import TensorboardLoggerConfig as TensorboardLoggerConfig
from .wandb import WandbLoggerConfig as WandbLoggerConfig

LoggerConfig: TypeAlias = Annotated[
    CSVLoggerConfig | TensorboardLoggerConfig | WandbLoggerConfig,
    C.Field(discriminator="name"),
]



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.callbacks.checkpoint.on_exception_checkpoint import OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig
from nshtrainer.callbacks.checkpoint.on_exception_checkpoint import CallbackConfigBase as CallbackConfigBase

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.loggers.tensorboard import TensorboardLoggerConfig as TensorboardLoggerConfig
from nshtrainer.loggers.tensorboard import BaseLoggerConfig as BaseLoggerConfig

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.loggers.csv import CSVLoggerConfig as CSVLoggerConfig
from nshtrainer.loggers.csv import BaseLoggerConfig as BaseLoggerConfig

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.nn.mlp import MLPConfig as MLPConfig
from nshtrainer.nn.mlp import BaseNonlinearityConfig as BaseNonlinearityConfig

# Type aliases
from nshtrainer.nn.mlp import NonlinearityConfig as NonlinearityConfig

# Submodule exports



from .base import LightningModuleBase as LightningModuleBase
from .config import BaseConfig as BaseConfig
from .config import DirectoryConfig as DirectoryConfig
from .config import MetricConfig as MetricConfig
from .config import TrainerConfig as TrainerConfig



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.metrics import MetricConfig as MetricConfig

# Type aliases

# Submodule exports
from . import _config as _config



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.model.base import EnvironmentConfig as EnvironmentConfig
from nshtrainer.model.base import BaseConfig as BaseConfig

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.callbacks.checkpoint.last_checkpoint import LastCheckpointCallbackConfig as LastCheckpointCallbackConfig
from nshtrainer.callbacks.checkpoint.last_checkpoint import CheckpointMetadata as CheckpointMetadata
from nshtrainer.callbacks.checkpoint.last_checkpoint import BaseCheckpointCallbackConfig as BaseCheckpointCallbackConfig

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.trainer.checkpoint_connector import CheckpointLoadingConfig as CheckpointLoadingConfig

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.callbacks.shared_parameters import SharedParametersConfig as SharedParametersConfig
from nshtrainer.callbacks.shared_parameters import CallbackConfigBase as CallbackConfigBase

# Type aliases

# Submodule exports



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.nn import MLPConfig as MLPConfig
from nshtrainer.nn import BaseNonlinearityConfig as BaseNonlinearityConfig
from nshtrainer.nn import TanhNonlinearityConfig as TanhNonlinearityConfig
from nshtrainer.nn import ReLUNonlinearityConfig as ReLUNonlinearityConfig
from nshtrainer.nn import PReLUConfig as PReLUConfig
from nshtrainer.nn import SiLUNonlinearityConfig as SiLUNonlinearityConfig
from nshtrainer.nn import LeakyReLUNonlinearityConfig as LeakyReLUNonlinearityConfig
from nshtrainer.nn import SoftsignNonlinearityConfig as SoftsignNonlinearityConfig
from nshtrainer.nn import MishNonlinearityConfig as MishNonlinearityConfig
from nshtrainer.nn import SigmoidNonlinearityConfig as SigmoidNonlinearityConfig
from nshtrainer.nn import SoftplusNonlinearityConfig as SoftplusNonlinearityConfig
from nshtrainer.nn.nonlinearity import SwiGLUNonlinearityConfig as SwiGLUNonlinearityConfig
from nshtrainer.nn import ELUNonlinearityConfig as ELUNonlinearityConfig
from nshtrainer.nn import SoftmaxNonlinearityConfig as SoftmaxNonlinearityConfig
from nshtrainer.nn import GELUNonlinearityConfig as GELUNonlinearityConfig
from nshtrainer.nn import SwishNonlinearityConfig as SwishNonlinearityConfig

# Type aliases
from nshtrainer.nn import NonlinearityConfig as NonlinearityConfig

# Submodule exports
from . import mlp as mlp
from . import nonlinearity as nonlinearity



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.callbacks.wandb_upload_code import WandbUploadCodeConfig as WandbUploadCodeConfig
from nshtrainer.callbacks.wandb_upload_code import CallbackConfigBase as CallbackConfigBase

# Type aliases

# Submodule exports



import copy
from collections.abc import Callable, Sequence
from typing import Literal, Protocol, runtime_checkable

import nshconfig as C
import torch
import torch.nn as nn
from typing_extensions import TypedDict, override

from .nonlinearity import BaseNonlinearityConfig, NonlinearityConfig


@runtime_checkable
class LinearModuleConstructor(Protocol):
    def __call__(
        self, in_features: int, out_features: int, bias: bool = True
    ) -> nn.Module: ...


class ResidualSequential(nn.Sequential):
    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return input + super().forward(input)


class MLPConfigDict(TypedDict):
    bias: bool
    """Whether to include bias terms in the linear layers."""

    no_bias_scalar: bool
    """Whether to exclude bias terms when the output dimension is 1."""

    nonlinearity: NonlinearityConfig | None
    """Activation function to use between layers."""

    ln: bool | Literal["pre", "post"]
    """Whether to apply layer normalization before or after the linear layers."""

    dropout: float | None
    """Dropout probability to apply between layers."""

    residual: bool
    """Whether to use residual connections between layers."""


class MLPConfig(C.Config):
    bias: bool = True
    """Whether to include bias terms in the linear layers."""

    no_bias_scalar: bool = True
    """Whether to exclude bias terms when the output dimension is 1."""

    nonlinearity: NonlinearityConfig | None = None
    """Activation function to use between layers."""

    ln: bool | Literal["pre", "post"] = False
    """Whether to apply layer normalization before or after the linear layers."""

    dropout: float | None = None
    """Dropout probability to apply between layers."""

    residual: bool = False
    """Whether to use residual connections between layers."""

    def to_kwargs(self) -> MLPConfigDict:
        return {
            "bias": self.bias,
            "no_bias_scalar": self.no_bias_scalar,
            "nonlinearity": self.nonlinearity,
            "ln": self.ln,
            "dropout": self.dropout,
            "residual": self.residual,
        }

    def create_module(
        self,
        dims: Sequence[int],
        pre_layers: Sequence[nn.Module] = [],
        post_layers: Sequence[nn.Module] = [],
        linear_cls: LinearModuleConstructor = nn.Linear,
    ):
        return MLP(
            dims,
            **self.to_kwargs(),
            pre_layers=pre_layers,
            post_layers=post_layers,
            linear_cls=linear_cls,
        )


def MLP(
    dims: Sequence[int],
    activation: BaseNonlinearityConfig
    | nn.Module
    | Callable[[], nn.Module]
    | None = None,
    nonlinearity: BaseNonlinearityConfig
    | nn.Module
    | Callable[[], nn.Module]
    | None = None,
    bias: bool = True,
    no_bias_scalar: bool = True,
    ln: bool | Literal["pre", "post"] = False,
    dropout: float | None = None,
    residual: bool = False,
    pre_layers: Sequence[nn.Module] = [],
    post_layers: Sequence[nn.Module] = [],
    linear_cls: LinearModuleConstructor = nn.Linear,
):
    """
    Constructs a multi-layer perceptron (MLP) with the given dimensions and activation function.

    Args:
        dims (Sequence[int]): List of integers representing the dimensions of the MLP.
        nonlinearity (Callable[[], nn.Module]): Activation function to use between layers.
        activation (Callable[[], nn.Module]): Activation function to use between layers.
        bias (bool, optional): Whether to include bias terms in the linear layers. Defaults to True.
        no_bias_scalar (bool, optional): Whether to exclude bias terms when the output dimension is 1. Defaults to True.
        ln (bool | Literal["pre", "post"], optional): Whether to apply layer normalization before or after the linear layers. Defaults to False.
        dropout (float | None, optional): Dropout probability to apply between layers. Defaults to None.
        residual (bool, optional): Whether to use residual connections between layers. Defaults to False.
        pre_layers (Sequence[nn.Module], optional): List of layers to insert before the linear layers. Defaults to [].
        post_layers (Sequence[nn.Module], optional): List of layers to insert after the linear layers. Defaults to [].

    Returns:
        nn.Sequential: The constructed MLP.
    """

    if activation is None:
        activation = nonlinearity

    if len(dims) < 2:
        raise ValueError("mlp requires at least 2 dimensions")
    if ln is True:
        ln = "pre"
    elif isinstance(ln, str) and ln not in ("pre", "post"):
        raise ValueError("ln must be a boolean or 'pre' or 'post'")

    layers: list[nn.Module] = []
    if ln == "pre":
        layers.append(nn.LayerNorm(dims[0]))

    layers.extend(pre_layers)

    for i in range(len(dims) - 1):
        in_features = dims[i]
        out_features = dims[i + 1]
        bias_ = bias and not (no_bias_scalar and out_features == 1)
        layers.append(linear_cls(in_features, out_features, bias=bias_))
        if dropout is not None:
            layers.append(nn.Dropout(dropout))
        if i < len(dims) - 2:
            match activation:
                case BaseNonlinearityConfig():
                    layers.append(activation.create_module())
                case nn.Module():
                    # In this case, we create a deep copy of the module to avoid sharing parameters (if any).
                    layers.append(copy.deepcopy(activation))
                case Callable():
                    layers.append(activation())
                case _:
                    raise ValueError(
                        "Either `nonlinearity` or `activation` must be provided"
                    )

    layers.extend(post_layers)

    if ln == "post":
        layers.append(nn.LayerNorm(dims[-1]))

    cls = ResidualSequential if residual else nn.Sequential
    return cls(*layers)



# fmt: off
# ruff: noqa
# type: ignore

__codegen__ = True

# Config classes
from nshtrainer.trainer._config import BestCheckpointCallbackConfig as BestCheckpointCallbackConfig
from nshtrainer.trainer._config import TensorboardLoggerConfig as TensorboardLoggerConfig
from nshtrainer.trainer._config import CSVLoggerConfig as CSVLoggerConfig
from nshtrainer.trainer._config import CheckpointSavingConfig as CheckpointSavingConfig
from nshtrainer.trainer._config import SharedParametersConfig as SharedParametersConfig
from nshtrainer.trainer._config import SanityCheckingConfig as SanityCheckingConfig
from nshtrainer.trainer._config import ReproducibilityConfig as ReproducibilityConfig
from nshtrainer.trainer._config import HuggingFaceHubConfig as HuggingFaceHubConfig
from nshtrainer.trainer._config import CheckpointLoadingConfig as CheckpointLoadingConfig
from nshtrainer.trainer._config import WandbLoggerConfig as WandbLoggerConfig
from nshtrainer.trainer._config import LoggingConfig as LoggingConfig
from nshtrainer.trainer._config import CallbackConfigBase as CallbackConfigBase
from nshtrainer.trainer._config import LastCheckpointCallbackConfig as LastCheckpointCallbackConfig
from nshtrainer.trainer._config import OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig
from nshtrainer.trainer._config import RLPSanityChecksConfig as RLPSanityChecksConfig
from nshtrainer.trainer._config import EarlyStoppingConfig as EarlyStoppingConfig
from nshtrainer.trainer._config import OptimizationConfig as OptimizationConfig
from nshtrainer.trainer._config import TrainerConfig as TrainerConfig
from nshtrainer.trainer._config import DebugFlagCallbackConfig as DebugFlagCallbackConfig
from nshtrainer.trainer._config import GradientClippingConfig as GradientClippingConfig

# Type aliases
from nshtrainer.trainer._config import CallbackConfig as CallbackConfig
from nshtrainer.trainer._config import CheckpointCallbackConfig as CheckpointCallbackConfig
from nshtrainer.trainer._config import LoggerConfig as LoggerConfig
from nshtrainer.trainer._config import ProfilerConfig as ProfilerConfig

# Submodule exports
